-------------------------------------------------------------------
-- Medusa
-- (macchinario esente da un significativo acronimo)
-- (C) Enrico Colombini 2009, 2014 - CC BY-SA 4.0
--
-- license: Attribution-ShareAlike 4.0 International
-- http://creativecommons.org/licenses/by-sa/4.0/
-------------------------------------------------------------------

local version = 'Medusa 1.5 - Copyright (C) 2009, 2014 Enrico Colombini - CC BY-SA 4.0'
local usage = 'usage: lua medusa.lua <book.lua> <book.html> [<config.cfg>]'

---------------------------------------------------------------------------
-- IMPORTANT NOTE:
--
-- this is not user-friendly code, it was written for my own use
-- there is no error checking - and I mean it
-- there is no documentation at all
-- there is no support
-- you have been warned

---------------------------------------------------------------------------
--[[

History
-------

- v1.0 (15 Feb 2009)

- v1.0.1 (27 May 2009)
  - fixed total #pages in report

- v1.1 (06 Jul 2009)
  - added PI()

- v1.2 (31 Dec 2009)
  - added currentPageReq
  - added bookinfo.stopPage
  - documented bookinfo
  - report: hidden pages with no generated instances
  - report: shown actual count of pages with generated instances
  - report: added data from bookinfo
  - report: shown month name to avoid possible day/month confusion

- v1.3 (04 Jan 2010)
  - added firstPage, stopPage override in config file
  - added warnings for firstPage redefined, stopPage defined
  - added compile time to report
  - added config.feedback

- v1.4 (08 Feb 2010)
  - removed auto-opening of paragraph
  - added version
  - added copyright
  - added <meta name="generator" ... >
  - added <meta name="copyright" ... >

- v1.5 (05 Sep 2014)
  - adapted to Lua 5.2: removed calls to setfenv, used _ENV
  - added _ENV also to explicit functions in Link, Choice
  - removed Continue()
  - removed invalid '%' characters from replacement string in pgf.T
  - removed degenerate O(n^2) case with many instances of same page
  - many small optimizations (also in memory usage)
  - uniformed terminology

--]]
---------------------------------------------------------------------------
--[[

  Theory of operation
  -------------------

- When an instance page is created, all its links are visited. The first
  instance page is created with an empty variable table (vtable).

- The page code can freely modify its variable table, as long as
  this is done deterministically from the received vtable values.

- For each link, the corresponding script (if any) is executed
  in an isolated environment, possibly modifying the local
  variable table (vtable) for that link only.

- If an instance page using the resulting combination of source page
  and vtable has already been created, it is referred to by the link.

- If such a page does not exist, a new instance page (source page + vtable)
  is scheduled for creation and referred to by the link.
  It is also added to the list of created instances.

- After page creation, pending instance page requests are examined: 
  for each request, a new instance page is created (using the source page 
  and vtable specified in the request).

- The debug flag adds readable information (page name, vtable)
  to the created book.

--]]
---------------------------------------------------------------------------
--[[

- Notes (waiting for documentation):
  - every source page is evaluated in a separate function
  - text must not contain <p>, </p>, <p/>
  - page title is <h3>, Title is <h4>, define in styles as needed
  - no page header if section title is empty, only sep. line if '-'
  - vtable must be shallow
  - beware of combinatorial explosion, reset unused variables to nil
  - to pass an empty vtable, use 'v = {}'
  - don't use random, must be fully deterministic
  - T must be inside P, use P"" to open a new paragraph

--]]


---------------------------------------------------------------------------
-- globals (actually file-scoped upvalues, not depending on current _ENV)
---------------------------------------------------------------------------

-- output file
local outfile = nil

-- configuration data, will be read from config file (if any)
local config = {}

-- book information (book.title, .author, , .copyright, .startPage, .stopPage)
local bookinfo = {}

-- table of source pages (keyed by page name) and their instances:
--   pagename = {
--       func=f,
--       [1], [2], ... -- instances of this page (as in requestedInstances below)
--       instances,    -- the same instances, keyed by their vstring
--   }
local pages = {}

-- list of requested instances to be created (LIFO for efficiency):
--   { pagename=name, instname=name__n, vtable=vtable, vstring=vtable_string }
local requestedInstances = {}

-- table for page generation functions (defined later)
local pgf = {}

-- state of current page processing
local pageState = {
    -- current environment for page function execution, used in link generation
    currentPageEnv = nil,
    -- instance page request currently being processed
    currentInstanceReq = nil,
    -- flag: a paragraph tag is currently open (HTML classes should replace this)
    paragraphIsOpen = false,
    -- flag: current paragraph is in italics (HTML classes should replace this)
    paragraphIsItalics = false,
}

-- function shortcuts for efficiency
local table_sort = table.sort
local table_concat = table.concat
local table_remove = table.remove


---------------------------------------------------------------------------
-- Medusa engine
---------------------------------------------------------------------------

function main()
    -- suppose all OK
    local result = 0

    -- get mandatory args
    local bookfname = arg[1]
    local outfname = arg[2]
    
    if not (bookfname and outfname) then
        io.write(version, '\n', usage)
        return 1 --> exit with error
    end

    -- read optional configuration data into config table
    local cfgfname = arg[3]
    if cfgfname then
        ReadConfiguration(cfgfname, config)
    end

    if config.feedback then
        io.write(version, '\n')
        io.write('compiling ', bookfname, ' --> ', outfname, ' using ', cfgfname, '\n')
    end

    -- read book source, apply macros
    local booksrc = ReadSourcePages(bookfname)
    -- compile into pages table, get book info table
    bookinfo = CompileSourcePages(booksrc, pages)

    -- override vars in bookinfo if defined in config, give warning
    if config.firstPage and (config.firstPage ~= bookinfo.firstPage) then
        io.write('-- warning: firstPage redefined in config file\n')
    end
    if config.stopPage or bookinfo.stopPage then
        io.write('-- warning: stopPage defined (either in book source or in config file)\n')
    end
    bookinfo.firstPage = config.firstPage or bookinfo.firstPage
    bookinfo.stopPage = config.stopPage or bookinfo.stopPage

    -- generate output book from firstPage according to config[] settings
    GenerateBook(outfname, bookinfo.firstPage)

    if config.feedback then
        io.write('compilation done at ', os.clock(), 's\n')
    end

    -- generate report as bookfname.report if requested
    if config.report then
        local basename = string.gsub(bookfname, '%.%w-$', '')
        if config.feedback then
            io.write('generating report --> ', basename, '.report\n')
        end
        GenerateReport(basename)
        if config.feedback then
            io.write('report done at ', os.clock(), 's\n')
        end
    end

    -- execute post-processing, if any
    local postProc = config.postProcessing
    if postProc then
        if config.feedback then
            io.write('post-processing\n')
        end
        -- post-processing can use iolib, oslib, stringlib, print
        config.io = io
        config.os = os
        config.string = string
        config.print = print
        result = postProc(outfname)
        if config.feedback then
            io.write('post-processing done at ', os.clock(), 's\n')
        end
    end

    if config.feedback then
        io.write('finished\n')
    end

    return result
end

---------------------------------------------------------------------------
-- read configuration data from given file into received cfg table

function ReadConfiguration(fname, cfg)
    -- read configuration file
    local f = io.open(fname, 'r') or error('cannot open configuration file: ' .. fname)
    local cfgsrc = f:read('*a')
    f:close()

    -- load (Lua compile) configuration inside config[] env
    local cfgcode = assert(load(cfgsrc, cfgsrc, 'bt', cfg))

    -- execute configuration code
    local r, err = pcall(cfgcode)
    -- exit if error in configuration
    r = r or error('error in reading configuration data:\n' .. err)
end

---------------------------------------------------------------------------
-- generate HTML output

function GenerateBook(outfname, startPage)
    -- open output file
    outfile = io.open(outfname, 'w') or error('cannot open output file: ' .. outfname)

    -- start HTML
    GenerateHtmlStart()

    -- request instance of first source page with empty vtable
    RequestInstancePage(startPage, {})

    -- serve instance page requests
    while #requestedInstances > 0 do
        -- get instance page description, remove from request list
        local instreq = table_remove(requestedInstances)
        -- generate instance page content, add requests for linked pages
        GenerateInstancePage(instreq)
    end

    -- end HTML
    GenerateHtmlEnd()

    -- done, close output file
    outfile:close()
end

---------------------------------------------------------------------------
-- read book source from fname, apply macros (do not alter line numbers),
-- return book source with macros applied

function ReadSourcePages(fname)
    local match = string.match
    local gsub = string.gsub
    local ssub = string.sub
    

    local insidePage = false
    local sectionTitle = nil
    -- table used to check for duplicated pages
    local pagesRead = {}

    -- read source file into table and apply line macros
    local t = {}
    for line in io.lines(fname) do
        -- page macro: "# name" --> "function page.name(_ENV)" ("end" later)
        local pagename = match(line, '^%#%s*([%w_]+)')
        -- section macro: '$section "title"' --> remember title
        local secTitle = match(line, '%$section%s*(%b[])')
        -- suppose line is unchanged
        local s = line
        if pagename then
            -- new page
            if match(pagename, '__') then
                error("'__' not allowed in page name: '" .. pagename .. "'")
            end
            -- check if already registered
            if pagesRead[pagename] then
                error('duplicated page: ' .. pagename)
            end
            
            pagesRead[pagename] = true
            s = ''
            -- close current page function, if any
            if insidePage then
                s = 'end; '
            end
            -- start new page function
            s = s .. 'function page.' .. pagename .. '(_ENV)'
            -- add section header, if any
            if sectionTitle then
                s = s .. ' PageHeader("' .. sectionTitle .. '")'
            end
            -- page function is open
            insidePage = true
        elseif secTitle then
            -- remember section title, if any
            sectionTitle = ssub(secTitle, 2, -2)
            s = '-- section: ' .. secTitle
        end
        
        -- in any case, add original line or replaced line
        -- (trying to remove comments is too risky because of multiline ones)
        t[#t + 1] = s
    end

    -- close last page function, if any
    if insidePage then
        t[#t + 1] = 'end\n' -- \n for test readability
        insidePage = false
    end

    -- convert table to newline-separated string
    local booksrc = table_concat(t, '\n')
    --print(booksrc); io.read() -- debug

    -- F macro: ", F{ ... }" --> ", function(_ENV) ... end" (may include balanced braces)
    local booksrc = gsub(booksrc, '%,%s*F%s*(%b{})',
                         function(s)
                             return ', function(_ENV) ' .. ssub(s, 2, -2) .. ' end'
                         end)

    -- add _ENV argument to explicitly declared functions in Link, Choice
    -- be sure not to add it to other page functions (it would mess up their args)
    booksrc = gsub(booksrc, '(Link%s*%([^%,%)]+,[^%,%)]+,%s*function%s*%()%s*(%))', '%1_ENV%2')
    booksrc = gsub(booksrc, '(Choice%s*%([^%,%)]+,[^%,%)]+,%s*function%s*%()%s*(%))', '%1_ENV%2')

    -- print(booksrc); io.read() -- debug

    return booksrc
end

---------------------------------------------------------------------------
-- compile book source into pgtable
-- return book info table

function CompileSourcePages(booksrc, pgtable)
    -- book environment (the 'book' inside it is the author-side info)
    local env = { page = {}, book = {} }

    -- load (Lua compile) book source inside env
    local bookcode = assert(load(booksrc, booksrc, 'bt', env))

    -- execute book code
    local r, err = pcall(bookcode)
    -- exit if error in source book
    r = r or error('error in reading source pages:\n' .. err)

    -- copy book info to bookinfo
    local infotable = CopyTable(env.book)

    -- copy page functions to pgtable, no instances yet
    for name, f in pairs(env.page) do
        pgtable[name] = { func=f, instances={} }
    end
    
    return infotable
end

---------------------------------------------------------------------------
-- request instance of given page with given vtable:
-- note: vtable must be a unique copy
--   if it already exists, do nothing
--   if it does not exist, add instance to both page and requestedInstances
-- return existing or new instance (description, not content)

function RequestInstancePage(pagename, vtable)
    -- get instance list for pagename page
    local page = pages[pagename]
    if not page then
        error(page, 'page not found: ' .. pagename)
    end
    
    -- get table of already defined instances for this page
    local instances = page.instances

    -- check if the requested instance already exists
    local vstring = VtableToString(vtable)
    local instance = instances[vstring]
    
    if not instance then
        -- the required instance does not already exist,
        -- create instance description with a new name
        local instname = pagename .. '__' .. (#page + 1)
        -- note: no need to copy vtable, it is already a copy
        instance = { pagename=pagename, instname=instname, vtable=vtable, vstring=vstring }
        -- add instance description to page instance list
        page[#page + 1] = instance
        -- reference this instance by its vstring
        instances[vstring] = instance
        -- also add instance description to list of requested pages
        -- (will be used to output the actual instance page content)
        requestedInstances[#requestedInstances + 1] = instance
    end

    -- existing or new (requested) instance
    return instance
end

---------------------------------------------------------------------------
-- process link request and script, request new instance page if needed
-- return new or existing instance page (description)

function ProcessLink(targetPage, script)
    -- execute script, if any, in a copy of the current environment
    -- to avoid altering its vtable (and possibly other vars)

    -- print(pageState.currentInstanceReq.pagename, ' --> ', targetPage) -- debug

    -- use a copy of the current environment containing only 1st level and vtable
    -- (CopyTable is shallow)
    local env = CopyTable(pageState.currentPageEnv)
    env.v = CopyTable(pageState.currentPageEnv.v)
    -- print('1 - env: ' .. VtableToString(env.v))

    if script then
        -- the script must be a (possibly macro-generated) function
        if type(script) ~= 'function' then
            error('page \'' .. targetPage .. '\': script must be a function')
        end

        -- execute it using env as environment table (first script arg is _ENV)
        local r, err = pcall(script, env)
        if not r then
            error('error in executing script for link to page ' .. targetPage .. ':\n' .. err)
        end
        -- print('1a - env: ' .. VtableToString(env.v))
    end
    -- print('2 - env: ' .. VtableToString(pageState.currentPageEnv.v))

    -- get and return (destination's) existing instance's name, or request new instance
    return RequestInstancePage(targetPage, env.v)
end

---------------------------------------------------------------------------
-- generate instance page contents into output file, 
-- add requests for linked instance pages
-- (see requestedInstances for instreq format, includes vtable)

function GenerateInstancePage(instreq)
    -- get page function
    local pagefunc = pages[instreq.pagename].func

    -- prepare execution environment with generation functions and vtable
    local env = CopyTable(pgf)
    env.v = instreq.vtable

    -- remember environment for current page function (used by links)
    pageState.currentPageEnv = env
    -- remember current instance page request
    pageState.currentInstanceReq = instreq

    -- check no open paragraph
    if pageState.paragraphIsOpen then
        error('paragraph should not be open in ' .. instreq.instname)
    end

    -- add pre-page, if any
    local prePage = config.prePage
    if prePage then
        outfile:write(prePage, '\n')
    end

    -- add anchor before page (the comment makes the <a> tag valid)
    outfile:write('<a name="', instreq.instname, '"><!-- --></a>\n')

    if config.debug then
        outfile:write(' <sub>[', instreq.instname, ', ', instreq.vstring, ']</sub> ')
    end

    -- execute page function in the prepared environment (first page arg is _ENV)
    -- (also request linked instance pages, if any)
    pagefunc(env)

    -- close any paragraph still open
    pgf.CloseP()

    -- add post-page, if any
    local postPage = config.postPage
    if postPage then
        outfile:write(postPage)
    end

    -- add empty line for readability
    outfile:write('\n')

    -- environment and current request are not needed anymore
    pageState.currentPageEnv = nil
    pageState.currentInstanceReq = nil
end

---------------------------------------------------------------------------
-- generate HTML document start, including header
-- with optional elements defined in configuration file
-- (HTML generation is primitive and should be restructured)

function GenerateHtmlStart(author, title)
    -- get optional configuration
    local doctype = config.doctype
    local encoding = config.encoding
    local style = config.style

    if doctype then
        outfile:write(doctype, '\n')
    end

    outfile:write('<html>\n')
    outfile:write('<head>\n')

    if encoding then
        outfile:write(encoding, '\n')
    end

    outfile:write('<meta name="author" content="', bookinfo.author, '">\n')
    if bookinfo.copyright then
        outfile:write('<meta name="copyright" content="', bookinfo.copyright, '">\n')
    end
    outfile:write('<meta name="generator" content="', version, '">\n')
    outfile:write('<title>', bookinfo.title, '</title>\n')

    if style then
        outfile:write(style)
    end

    outfile:write('</head>\n')
    outfile:write('<body>\n\n')
end

---------------------------------------------------------------------------
-- generate HTML document end
-- (HTML generation is primitive and should be restructured)

function GenerateHtmlEnd(author, title)
    outfile:write('</body>\n')
    outfile:write('</html>\n')
end


---------------------------------------------------------------------------
-- output (HTML) page generation functions (called by page function)
-- note: very primitive, ad-hoc code (should replace with better API)
---------------------------------------------------------------------------

-- PageHeader(title)  (reserved, automatic) generate page header
-- T(text, ...)       normal text
-- P(text, ...)       open paragraph, shortcut for T("<p>", text, ...)
-- PI(text, ...)      open italics paragraph, shortcut for T("<p><i>", text, ...)
-- CloseP()           close paragraph if open (automatic when needed)
-- Title(text)        generic <h4> title
-- N(n)               n newlines (default = 1), same as T"<br>"
--
-- Link(text, page [, script])    link within text flow (script is a function)
-- Choice(text, page [, script])  link on separate line (script is a function)
--
-- note: do not call local functions from Link/Choice scripts

---------------------------------------------------------------------------
-- output page header with section title (if not empty)

function pgf.PageHeader(title)
    -- special case: section title '-' means separator only
    if title == '-' then
        -- use non-breakable space to force correct layout
        title = '&nbsp;'
    end
    -- if no title at all, no separating line either
    if title ~= '' then
        outfile:write('<h3>', title, '</h3>\n') -- \n added for readability
        -- separator
        outfile:write('<hr>\n') -- \n added for readability
    end
end

---------------------------------------------------------------------------
-- generate HTML text made of 1 or more strings

function pgf.T(...)
    outfile:write(...)
end

---------------------------------------------------------------------------
-- start normal paragraph with 1 or more strings (do not close it)

function pgf.P(...)
    -- close previous paragraph, if open
    pgf.CloseP()
    -- start new paragraph (will be auto-closed)
    outfile:write('<p>')
    pageState.paragraphIsOpen = true
    pageState.paragraphIsItalics = false
    -- output text
    outfile:write(...)
end

---------------------------------------------------------------------------
-- start italics paragraph with 1 or more strings (do not close it)

function pgf.PI(...)
    -- close previous paragraph, if open
    pgf.CloseP()
    -- start new paragraph (will be auto-closed)
    outfile:write('<p><i>')
    pageState.paragraphIsOpen = true
    pageState.paragraphIsItalics = true
    -- output text
    outfile:write(...)
end

---------------------------------------------------------------------------
-- close paragraph if currently open (else do nothing)

function pgf.CloseP()
    -- close italics, if open
    if pageState.paragraphIsItalics then
        outfile:write('</i>')
        pageState.paragraphIsItalics = false
    end
    if pageState.paragraphIsOpen then
        outfile:write('</p>\n') -- \n added for readability
        pageState.paragraphIsOpen = false
    end
end

---------------------------------------------------------------------------
-- generate <h4> title

function pgf.Title(txt)
    outfile:write('<h4>', txt, '</h4>\n') -- \n added for readability
end

---------------------------------------------------------------------------
-- generate n newlines, default = 1 (use \n for newlines inside text)

function pgf.N(n)
    n = n or 1
    for i = 1, n do
        outfile:write('<br>')
    end
    outfile:write('\n') -- \n added for readability
end

---------------------------------------------------------------------------
-- generate link within text flow,
-- execute optional script before page change

function pgf.Link(text, page, script)
    if pageState.currentInstanceReq.pagename == bookinfo.stopPage then
        -- stop page reached, output underlined text instead of link
        outfile:write('<u>', text, '</u>')
    else
        -- normal link
        local instance = ProcessLink(page, script)

        outfile:write('<a href="#', instance.instname, '">', text, '</a>')

        if config.debug then
            outfile:write(' <sub>[', instance.instname, ', ', 
                          instance.vstring, ']</sub> ')
        end
    end
end

---------------------------------------------------------------------------
-- generate choice on separate line
-- execute optional script before page change

function pgf.Choice(text, page, script)
    -- close previous paragraph, if open
    pgf.CloseP()
    -- generate link in list element
    outfile:write('<ul><li>')
    pgf.Link(text, page, script)
    outfile:write('</li></ul>\n') -- \n added for readability
end


---------------------------------------------------------------------------
-- Utilities
---------------------------------------------------------------------------

-- return a shallow copy of a table

function CopyTable(t)
    local tc = {}
    for k, v in pairs(t) do
        tc[k] = v
    end
    return tc
end

---------------------------------------------------------------------------
-- make a shallow compare of two tables,
-- return true if they have the same keys and values

function CompareTables(t1, t2)
    -- all key-value pairs in t1 must be in t2
    for k, v in pairs(t1) do
        if t2[k] ~= v then return false end
    end
    -- there must not be other keys in t2
    for k, v in pairs(t2) do
        if t1[k] == nil then return false end
    end
    return true
end

---------------------------------------------------------------------------
-- return a string representing the first level of a vtable
-- in the format: { k=v, k=v, ... } with the keys in alphabetical order,
-- strings are surrounded by single quotes, special chars are not escaped

function VtableToString(vt)
    -- get sorted list of keys
    local keys = {}
    for k, v in pairs(vt) do
        keys[#keys + 1] = k
    end
    table_sort(keys)
    
    -- accumulate into a table for efficiency
    local t = {}
    t[#t + 1] = '{ '
    for _, k in ipairs(keys) do
        local v = vt[k]
        -- this is faster than using string.format
        -- (string escaping is not needed here)
        t[#t + 1] = tostring(k)
        t[#t + 1] = '='
        if type(v) == 'string' then
            -- safety catch in case of both string and non-string with equal value
            t[#t + 1] = "'"
            t[#t + 1] = tostring(v)
            t[#t + 1] = "'"
        else
            t[#t + 1] = tostring(v)
        end
        t[#t + 1] = ', '
    end
    t[#t + 1] = '}'

    local r = table_concat(t)
    if r == '{ }' then r = '{}' end -- cosmetic
    return r
end


---------------------------------------------------------------------------
-- Reporting
---------------------------------------------------------------------------

function GenerateReport(basename)
    local compileTime = os.clock()

    local format = string.format
    local sep = string.rep('-', 76) .. '\n'

    -- prepare table of page names, count total number of used pages 
    -- and number of created instance pages
    local pagenames = {}
    local numUsedPages = 0
    local numInstances = 0
    for k, _ in pairs(pages) do
        pagenames[#pagenames + 1] = k
        local n = #pages[k]
        if n > 0 then
            numUsedPages = numUsedPages + 1
            numInstances = numInstances + n
        end
    end

    -- create report file
    local f = io.open(basename .. '.report', 'w')
    f:write(sep)
    f:write("-- Report for '", basename, "' - ", os.date('%d %b %Y %H:%M'), "\n")
    f:write(sep, '\n')

    -- write summary
    f:write('-- Summary:\n\n')
    f:write('title:          ', bookinfo.title, '\n')
    f:write('author:         ', bookinfo.author, '\n')
    f:write('\n')
    f:write('first page:     ', bookinfo.firstPage, '\n')
    if bookinfo.stopPage then
        f:write('stop page:      ', bookinfo.stopPage, '\n')
    end
    f:write('source pages:   ', numUsedPages, '/', #pagenames, '\n')
    f:write('created pages:  ', numInstances, '\n')
    f:write('\n')
    f:write('compilation time: ', compileTime, ' s\n')
    f:write('\n')

    -- list pages sorted by name
    table_sort(pagenames)
    f:write(sep)
    f:write('-- Pages by name:\n\n')
    for _, name in ipairs(pagenames) do
        local n = #pages[name]
        -- only if the page has instances
        if n > 0 then
            local s = format('%5d  %s\n', n, name)
            f:write(s)
        end
    end
    f:write('\n')

    -- list pages sorted by number of instances
    table_sort(pagenames,
               function(page1, page2) return #pages[page1] > #pages[page2] end)
    f:write(sep)
    f:write('-- Pages by number of created instances:\n\n')
    for _, name in ipairs(pagenames) do
        local n = #pages[name]
        -- only if the page has instances
        if n > 0 then
            local s = format('%5d  %s\n', n, name)
            f:write(s)
        end
    end
    f:write('\n')

    -- list instance pages by page name, with their vtable
    table_sort(pagenames)
    f:write(sep)
    f:write('-- Instance pages by page name:\n\n')
    for _, name in ipairs(pagenames) do
        local page = pages[name]
        -- only if the page has instances
        if #page > 0 then
            --local s = format('%5d  %s\n', #page, name)
            f:write(name, '\n')
            -- instances of this page
            for i, instance in ipairs(page) do
                f:write('   ', instance.instname, '  ', instance.vstring, '\n')
            end
            f:write('\n')
        end
    end

    -- done, close report file
    f:write(sep)
    f:close()
end


---------------------------------------------------------------------------
-- Program start
---------------------------------------------------------------------------

os.exit(main())

---------------------------------------------------------------------------
